@@ -1,25 +1,33 @@
-"""
-This program downloads / scrapes Apple maps for free.
-OPEN SOURCED UNDER GPL-V3.0.
-Author : Jimut Bahan Pal | jimutbahanpal@yahoo.com
-"""
-#pylint: disable= global-statement
-#cSpell:words imghdr, tqdm, asinh, jimut, bahan
-# imports
+
+# ========================================================
+# This program fetches tiles from satellites.pro for free.
+# OPEN SOURCED UNDER GPL-V3.0.
+# Author : Jimut Bahan Pal | jimutbahanpal@yahoo.com
+# Project Website: https://github.com/Jimut123/jimutmap
+# pylint: disable = global-statement
+# cSpell: words imghdr, tqdm, asinh, jimut, bahan
+# ========================================================
 
 import ssl
 import os
 import time
 import math
-from os.path import join, exists, normpath, relpath
 import imghdr
-from multiprocessing.pool import ThreadPool
-from typing import Tuple
-from tqdm import tqdm
-import numpy as np
 import requests
+import numpy as np
+import datetime as dt
+from tqdm import tqdm
+import multiprocessing
+from typing import Tuple
 from selenium import webdriver
 import chromedriver_autoinstaller
+from multiprocessing.pool import ThreadPool
+from os.path import join, exists, normpath, relpath
+
+
+
+
+
 
 # Ignore SSL certificate errors
 ctx = ssl.create_default_context()
@@ -30,6 +38,10 @@ headers = {
     'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:21.0) Gecko/20100101 Firefox/21.0'
 }
 
+
+
+
+
 # To synchronize
 
 LOCK_VAR = 0
@@ -61,7 +73,7 @@ class api:
             Helpful debugging output
 
         threads_: int (default= 4)
-            Thread limit for process. Max 50
+            Thread limit for process. Max depending on CPU cores
 
         container_dir:str (default= "")
             When downloading images, place them in this directory.
@@ -77,11 +89,21 @@ class api:
         self.set_bounds(min_lat_deg, max_lat_deg, min_lon_deg, max_lon_deg)
         self.zoom = zoom
         self.verbose = bool(verbose)
+        
+        # To get the maximum number of threads
+        MAX_CORES = multiprocessing.cpu_count()
+        if threads_> MAX_CORES:
+            print("Sorry, {} -- threads unavailable, using maximum CPU threads : {}".format(threads_,MAX_CORES))
+            threads_ = MAX_CORES
+        
         LOCKING_LIMIT = threads_
+
         if self.verbose:
             print(self.ac_key,self.min_lat_deg,self.max_lat_deg,self.min_lon_deg,self.max_lon_deg,self.zoom,self.verbose,LOCKING_LIMIT)
         self._getMasks = True
         self.container_dir = container_dir
+        print("Initializing jimutmap ... Please wait...")
+        
 
     @property
     def container_dir(self) -> str:
@@ -154,7 +176,7 @@ class api:
             if newACKey is not None:
                 raise ValueError("Invalid AccessKey string")
 
-    def _getAPIKey(self) -> str:
+    def _getAPIKey(self, timeout:float= 60) -> str:
         """
         Use a headless Chrome/Chromium instance to scrape the access key
         from the data-map-printing-background attribute of Apple Maps.
@@ -166,15 +188,23 @@ class api:
         options.add_argument('headless')
         driver = webdriver.Chrome(executable_path= chromeDriverPath, options=options)
         driver.get("https://satellites.pro/USA_map#37.405074,-94.284668,5")
-        time.sleep(15)
-        baseMap = driver.find_element_by_css_selector("#map-canvas .leaflet-mapkit-mutant")
-        mapData = baseMap.get_attribute("data-map-printing-background")
-        accessKeyStart = mapData.find(KEY_START)
-        accessKeyEnd = accessKeyStart + int(1.5 * len(SAMPLE_KEY))
-        searchForKey = mapData[accessKeyStart:accessKeyEnd]
-        keyContents = searchForKey[len(KEY_START):]
-        keyEnd = keyContents.find("&")
-        keyContents = keyContents[:keyEnd]
+        keyContents = None
+        loopStartTime = dt.datetime.now()
+        while keyContents is None and (dt.datetime.now() - loopStartTime).total_seconds() < timeout:
+            time.sleep(2.5)
+            try:
+                baseMap = driver.find_element_by_css_selector("#map-canvas .leaflet-mapkit-mutant")
+                mapData = baseMap.get_attribute("data-map-printing-background")
+                accessKeyStart = mapData.find(KEY_START)
+                accessKeyEnd = accessKeyStart + int(1.5 * len(SAMPLE_KEY))
+                searchForKey = mapData[accessKeyStart:accessKeyEnd]
+                keyContents = searchForKey[len(KEY_START):]
+                keyEnd = keyContents.find("&")
+                keyContents = keyContents[:keyEnd]
+            except Exception: #pylint: disable= broad-except
+                keyContents = None
+        if keyContents is None:
+            raise TimeoutError(f"Unable to automatically fetch API key in {timeout}s")
         self.ac_key = keyContents
         return keyContents
 
@@ -272,6 +302,7 @@ class api:
         except AssertionError:
             try:
                 # get the image tile and the mask tile for the same
+                # sample sat tile: https://sat-cdn2.apple-mapkit.com/tile?style=7&size=1&scale=1&z=19&x=390843&y=228270&v=9262&accessKey=1649243787_2102081627305478489_%2F_hIz9LjsZkMj6NE7y%2BimXS9vFQbxfjLBClZR7yqyFtsE%3D&emphasis=standard&tint=light
                 req_url = f"https://sat-cdn1.apple-mapkit.com/tile?style=7&size=1&scale=1&z={self.zoom}&x={xTile}&y={yTile}&v={vNumber}{self.ac_key}"
                 if self.verbose:
                     print(req_url)
@@ -287,7 +318,7 @@ class api:
                     pass
                 with open(file_name, 'wb') as fh:
                     fh.write(r.content)
-                if imghdr.what(file_name) is 'jpeg':
+                if imghdr.what(file_name) == 'jpeg':
                     if self.verbose:
                         print(file_name,"JPEG")
                 else:
@@ -299,12 +330,21 @@ class api:
                     print(e)
         if getMask:
             ext = file_name.split('.').pop()
-            file_name_road = file_name[:-len(ext)]+"_road.png"
+            file_name_road = file_name[:-len(ext)-1]+"_road.png"
             try:
                 assert exists(file_name_road)
             except AssertionError:
                 for cdnLevel in range(1, 5):
-                    req_url = f"https://cdn{cdnLevel}.apple-mapkit.com/ti/tile?country=US&region=US&style=46&size=1&x={xTile}&y={yTile}&z={self.zoom}&scale=1&lang=en&v=2008184&poi=0{self.ac_key}&labels=0"
+                    # change the 2nd auth according to the datetime for downloading the roads masks
+                    today_date = str(dt.date.today())
+                    year = today_date[0:4]
+                    month = today_date[5:7]
+                    day = today_date[8:10]
+                    env_key = str(year)+str(month)+str(day)
+                    # sample road tile 1: https://cdn3.apple-mapkit.com/ti/tile?country=IN&region=IN&style=46&size=1&x=390842&y=228268&z=19&scale=1&lang=en&v=2204054&poi=1&accessKey=1649243787_2102081627305478489_%2F_hIz9LjsZkMj6NE7y%2BimXS9vFQbxfjLBClZR7yqyFtsE%3D&emphasis=standard&tint=light
+                    # sample road tile 2: https://cdn4.apple-mapkit.com/ti/tile?country=IN&region=IN&style=46&size=1&x=296223&y=176608&z=19&scale=1&lang=en&v=2204054&poi=1&accessKey=1649243787_2102081627305478489_%2F_hIz9LjsZkMj6NE7y%2BimXS9vFQbxfjLBClZR7yqyFtsE%3D&emphasis=standard&tint=light
+                    
+                    req_url = f"https://cdn{cdnLevel}.apple-mapkit.com/ti/tile?country=US&region=US&style=46&size=1&x={xTile}&y={yTile}&z={self.zoom}&scale=1&lang=en&v={env_key}4&poi=1{self.ac_key}&emphasis=standard&tint=light"
                     try:
                         # image and mask retrieval
                         # For the roads data
@@ -313,7 +353,7 @@ class api:
                         r = requests.get(req_url, headers= headers)
                         with open(file_name_road, 'wb') as fh:
                             fh.write(r.content)
-                        if imghdr.what(file_name_road) is 'png':
+                        if imghdr.what(file_name_road) == 'png':
                             if self.verbose:
                                 print(file_name_road,"PNG")
                             break # Success
@@ -348,9 +388,9 @@ class api:
             # results and we'll fetch nothing
             raise ValueError(f"Latitude and longitude bounds must be separated by at least the latLonResolution (currently {latLonResolution}). Either shrink the resolution value or increase the separation of your minimum/maximum latitude and longitude.")
 
+        URL_ALL = []
         for i in tqdm(np.arange(min_lat, max_lat, latLonResolution)):
             tp = None
-            URL_ALL = []
             for j in np.arange(min_lon, max_lon, latLonResolution):
                 URL_ALL.append(self.make_url(i,j))
             if self.verbose:
@@ -371,3 +411,5 @@ class api:
                 # process has cleared. As a practical matter, this will
                 # clear _several_ threads and keep up performance
                 tp.join()
+
+        
